# -*- coding: utf-8 -*-
"""
Hilltop read functions.
"""
try:
    import Hilltop
except:
    print('The Hilltop module could not load. Change the python path to where the Hilltop.pyd resides.')
import pandas as pd
from hilltoppy.util import time_switch

#####################################################
#### New method - might have issues


def site_list(hts):
    """
    Function to return a list of sites from an hts file. Exists for consistancy.

    Parameters
    ----------
    hts : str
        Path to hts file.

    Returns
    -------
    list
    """
    dfile1 = Hilltop.Connect(hts)
    site_list = Hilltop.SiteList(dfile1)
    Hilltop.Disconnect(dfile1)

    return site_list


def measurement_list(hts, sites=None):
    """
    Function to read all of the sites in an hts file and the associated site info.

    Parameters
    ----------
    hts : str
        Path to hts file.
    sites : list
        A list of site names to return.

    Returns
    -------
    DataFrame
    """

    dfile1 = Hilltop.Connect(hts)
    site_list = Hilltop.SiteList(dfile1)

    if not site_list:
        print('No sites in ' + hts)
        return None

    if isinstance(sites, list):
        site_list = [i for i in site_list if i in sites]
        if not site_list:
            print('No sites in ' + hts)
            return None

    site_info_list = []

    for i in site_list:
        try:
            info1 = Hilltop.MeasurementList(dfile1, i)
        except SystemError as err:
            print('Site ' + str(i) + " didn't work. Error: " + str(err))
            continue
        info1.loc[:, 'site'] = i.encode('ascii', 'ignore').decode()
        site_info_list.append(info1)
    site_info = pd.concat(site_info_list)

    site_info.loc[:, 'Start Time'] = pd.to_datetime(site_info.loc[:, 'Start Time'], format='%d-%b-%Y %H:%M:%S', errors='coerce')
    site_info.loc[:, 'End Time'] = pd.to_datetime(site_info.loc[:, 'End Time'], format='%d-%b-%Y %H:%M:%S', errors='coerce')

    bad_sites = site_info[site_info['Start Time'].isnull() | site_info['End Time'].isnull() | (site_info['Start Time'] < '1900-01-01') | (site_info['End Time'] > pd.Timestamp.today())]

    if not bad_sites.empty:
        print('There are ' + str(len(bad_sites)) + ' sites with bad times')
        print(bad_sites)
        site_info = site_info[~(site_info['Start Time'].isnull() | site_info['End Time'].isnull() | (site_info['Start Time'] < '1900-01-01') | (site_info['End Time'] > pd.Timestamp.today()))]

    len_all = len(site_list)
    len_got = len(site_info.site.unique())
    print('Missing ' + str(len_all - len_got) + ' sites, which is ' + str(round(100 * ((len_all - len_got)/len_all), 1)) + '% of the total')

    Hilltop.Disconnect(dfile1)
    return site_info


def get_data(hts, sites=None, mtypes=None, from_date=None, to_date=None, agg_method='Average', agg_n=1, agg_period='day', output_missing_sites=False, site_info=None):
    """
    Function to read time series from an hts file.

    Parameters
    ----------
    hts : str
        Path to the hts file.
    sites : list
        A list of site names within the hts file.
    mtypes : list
        A list of measurement types that should be returned.
    from_date : str
        The start date to retreive from the data in ISO format (e.g. '2011-11-30 00:00').
    to_date : str
        The end date to retreive from the data in ISO format (e.g. '2011-11-30 00:00').
    agg_method : str
        Options are '', 'Interpolate', 'Average', 'Total', 'Moving Average', and 'EP'.
    agg_period : str
        The resample period (e.g. 'day', 'month').
    agg_n : int
        The number of periods (e.g. 1 for 1 day).
    output_site_data : bool
        Should the sites data be output?
    sites_info : DataFrame
        The DataFrame return from the get_sites_mtypes function. If this is passed than get_sites_mtypes is not run.

    Returns
    -------
    DataFrame
    """

    if not isinstance(site_info, pd.DataFrame):
        site_info = measurement_list(hts, sites=sites)

    if isinstance(mtypes, list):
        site_info = site_info[site_info['Measurement'].isin(mtypes)]

    if isinstance(from_date, str):
        from_date1 = pd.to_datetime(from_date)
        site_info = site_info[site_info['End Time'] >= from_date1].copy()
        site_info.loc[site_info['Start Time'] < from_date1, 'Start Time'] = from_date1

    if isinstance(to_date, str):
        to_date1 = pd.to_datetime(to_date)
        site_info = site_info[site_info['Start Time'] <= to_date1].copy()
        site_info.loc[site_info['End Time'] > to_date1, 'End Time'] = to_date1

    ## Convert datetime and interval to hilltop specific formats
    pd_time_code = time_switch(agg_period)
    site_info.loc[:, 'Start Time'] = site_info.loc[:, 'Start Time'].dt.ceil(str(agg_n) + pd_time_code).dt.strftime('%d-%b-%Y %H:%M:%S')
    site_info.loc[:, 'End Time'] = site_info.loc[:, 'End Time'].dt.strftime('%d-%b-%Y %H:%M:%S')

    ht_interval = str(agg_n) + ' ' + agg_period

    ## Extract the ts data
    data1 = []

    dfile1 = Hilltop.Connect(hts)
    for i in site_info.index:
        site = site_info.loc[i, 'site']
        mtype = site_info.loc[i, 'Measurement']
        start = site_info.loc[i, 'Start Time']
        end = site_info.loc[i, 'End Time']

        try:
            d1 = Hilltop.GetData(dfile1, site, mtype, start, end, method=agg_method, interval=ht_interval, alignment='00:00')
        except Exception as err:
            err1 = err
            print(err1)
            print('Extraction failed for site ' + str(site) + ' and mtype ' + str(mtype))
            continue

        if d1.empty:
            print('No data for site ' + str(site) + ' and mtype ' + str(mtype))
            continue

        if (pd_time_code == 'D'):
            d1.index = d1.index.normalize()
        d1.name = 'data'
        d1.index.name = 'time'
        d2 = d1.reset_index()
        d2.loc[:, 'site'] = site
        d2.loc[:, 'mtype'] = mtype

        data1.append(d2)

    try:
        data2 = pd.concat(data1).set_index(['mtype', 'site', 'time']).data
    except MemoryError:
        print('Not enough RAM')

    return data2




# -*- coding: utf-8 -*-
"""
Hilltop read functions for COM connection.
Hilltop uses a fixed base date as 1940-01-01, while the standard unix/POSIT base date is 1970-01-01.
"""
try:
    from win32com.client import Dispatch, pywintypes, makepy
except:
    print('Install pywin32 or com functions will not work')
from pandas import concat, to_datetime, to_numeric, DataFrame, merge
from hilltoppy.util import pytime_to_datetime, time_switch

######################################################
#### COM access method


def makepy_hilltop(hlib='Hilltop Data Access'):
    """
    Function to generate the Hilltop COM module.

    Parameters
    ----------
    hlib : str
        The name of the COM library.

    Returns
    -------
    None
    """
    makepy.GenerateFromTypeLibSpec(hlib, verboseLevel=1)


def measurement_list(hts, sites=None, mtypes=None, rem_wq_sample=True):
    """
    Function to read the site names, measurement types, and units of a Hilltop hts file. Returns a DataFrame.

    Parameters
    ----------
    hts : str
        Path to the hts file.
    sites : list or None
        A list of site names within the hts file.
    mtypes : list or None
        A list of measurement types that should be returned.
    rem_wq_sample : bool
        In Hilltop 'WQ Sample' is a measurement type placemarker for the additional sample data. It doesn't generally apply when querying for the combo of sites/measurement types. True removes this instance from the returned DataFrame.

    Returns
    -------
    DataFrame
    """

    cat = Dispatch("Hilltop.Catalogue")
    if not cat.Open(hts):
        raise ValueError(cat.errmsg)

    dfile = Dispatch("Hilltop.DataRetrieval")
    try:
        dfile.Open(hts)
    except ValueError:
        print(dfile.errmsg)

    sites_lst = []

    ### Iterate through all sites/datasources/mtypes
    cat.StartSiteEnum
    while cat.GetNextSite:
        site_name = str(cat.SiteName.encode('ascii', 'ignore').decode())
        if sites is None:
            pass
        elif site_name in sites:
            pass
        else:
            continue
        while cat.GetNextDataSource:
            ds_name = str(cat.DataSource.encode('ascii', 'ignore').decode())
            try:
                start1 = pytime_to_datetime(cat.DataStartTime)
                end1 = pytime_to_datetime(cat.DataEndTime)
            except ValueError:
                bool_site = dfile.FromSite(site_name, ds_name, 1)
                if bool_site:
                    start1 = pytime_to_datetime(cat.DataStartTime)
                    end1 = pytime_to_datetime(cat.DataEndTime)
                else:
                    print('No site data for ' + site_name + '...for some reason...')
            while cat.GetNextMeasurement:
                mtype1 = str(cat.Measurement.encode('ascii', 'ignore').decode())
                if mtype1 == 'Item2':
                    continue
                elif mtypes is None:
                    pass
                elif mtype1 in mtypes:
                    pass
                else:
                    continue
                divisor = cat.Divisor
                unit1 = str(cat.Units.encode('ascii', 'ignore').decode())
                if unit1 == '%':
#                    print('Site ' + name1 + ' has no units')
                    unit1 = ''
                sites_lst.append([site_name, ds_name, mtype1, unit1, divisor, str(start1), str(end1)])

    sites_df = DataFrame(sites_lst, columns=['site', 'data_source', 'mtype', 'unit', 'divisor', 'start_date', 'end_date'])
    if rem_wq_sample:
        sites_df = sites_df[~(sites_df.mtype == 'WQ Sample')]
    dfile.Close()
    cat.Close()
    return sites_df


def get_data_quantity(hts, sites=None, mtypes=None, start=None, end=None, agg_period=None, agg_n=1, fun=None, output_site_data=False, exclude_mtype=None, sites_df=None):
    """
    Function to read water quantity data from an hts file.

    Parameters
    ----------
    hts : str
        Path to the hts file.
    sites : list
        A list of site names within the hts file.
    mtypes : list
        A list of measurement types that should be returned.
    start : str
        The start date to retreive from the data in ISO format (e.g. '2011-11-30 00:00').
    end : str
        The end date to retreive from the data in ISO format (e.g. '2011-11-30 00:00').
    agg_period : str
        The resample period (e.g. 'day', 'month').
    agg_n : int
        The number of periods (e.g. 1 for 1 day).
    fun : str
        The resampling function.
    output_site_data : bool
        Should the sites data be output?
    sites_df : DataFrame
        The DataFrame return from the rd_hilltop_sites function. If this is passed than rd_hilltop_sites is not run.

    Returns
    -------
    DataFrame
    """

    agg_name_dict = {'sum': 4, 'count': 5, 'mean': 1}
    agg_unit_dict = {'l/s': 1, 'm3/s': 1, 'm3/hour': 1, 'mm': 1, 'm3': 4}
    unit_convert = {'l/s': 0.001, 'm3/s': 1, 'm3/hour': 1, 'mm': 1, 'm3': 1}

    ### First read all of the sites in the hts file and select the ones to be read
    if not isinstance(sites_df, DataFrame):
        sites_df = measurement_list(hts, sites=sites, mtypes=mtypes)
    sites_df = sites_df[sites_df.unit.isin(list(agg_unit_dict.keys()))]
    if isinstance(exclude_mtype, list):
        sites_df = sites_df[~sites_df.mtype.isin(exclude_mtype)]

    ### Select out the sites/mtypes within the date range
    if isinstance(start, str):
        sites_df = sites_df[sites_df.end_date >= start]
    if isinstance(end, str):
        sites_df = sites_df[sites_df.start_date <= end]

    ### Open the hts file
    dfile = Dispatch("Hilltop.DataRetrieval")
    try:
        dfile.Open(hts)
    except ValueError:
        print(dfile.errmsg)

    ### Iterate through he hts file
    df_lst = []
    for i in sites_df.index:
        site = sites_df.loc[i, 'site']
        mtype = sites_df.loc[i, 'mtype']
        unit = sites_df.loc[i, 'unit']
        if fun is None:
            agg_val = agg_unit_dict[unit]
        else:
            agg_val = agg_name_dict[fun]
        if dfile.FromSite(site, mtype, 1):

            ## Set up start and end times and aggregation initiation
            start_time = pytime_to_datetime(dfile.DataStartTime)
            end_time = pytime_to_datetime(dfile.DataEndTime)
            if (start_time.year < 1900) | (end_time.year < 1900):
                print('Site ' + site + ' has a start or end time prior to 1900')
                continue
            if (start is None):
                if (agg_period is not None):
                    start1 = str(to_datetime(start_time).ceil(str(agg_n) + time_switch(agg_period)))
                else:
                    start1 = dfile.DataStartTime
            else:
                start1 = start
            if end is None:
                end1 = dfile.DataEndTime
            else:
                end1 = end
            if not dfile.FromTimeRange(start1, end1):
                continue
            if (agg_period is not None):
                dfile.SetMode(agg_val, str(agg_n) + ' ' + agg_period)

            ## Extract data
            data = []
            time = []
            if dfile.getsinglevbs == 0:
                t1 = dfile.value
                if isinstance(t1, str):
                    print('site ' + site + ' has nonsense data')
                else:
                    data.append(t1)
                    time.append(str(pytime_to_datetime(dfile.time)))
                    while dfile.getsinglevbs != 2:
                        data.append(dfile.value)
                        time.append(str(pytime_to_datetime(dfile.time)))
                    if data:
                        df_temp = DataFrame({'time': time, 'data': data, 'site': site, 'mtype': mtype})
                        df_lst.append(df_temp)

    dfile.Close()
    if df_lst:
        df1 = concat(df_lst)
        df1.loc[:, 'time'] = to_datetime(df1.loc[:, 'time'])
        df2 = df1.set_index(['mtype', 'site', 'time']).data * unit_convert[unit]
    else:
        df2 = DataFrame([], index=['mtype', 'site', 'time'])
    if output_site_data:
        return df2, sites_df
    else:
        return df2


def get_data_quality(hts, sites=None, mtypes=None, start=None, end=None, dtl_method=None, output_site_data=False, mtype_params=None, sample_params=None, sites_df=None):
    """
    Function to read water quality data from an hts file.

    Parameters
    ----------
    hts : str
        Path to the hts file.
    sites : list
        A list of site names within the hts file.
    mtypes : list
        A list of measurement types that should be returned.
    start : str
        The start date to retreive from the data in ISO format (e.g. '2011-11-30 00:00').
    end : str
        The end date to retreive from the data in ISO format (e.g. '2011-11-30 00:00').
    dtl_method : None, 'standard', 'trend'
        The method to use to convert values under a detection limit to numeric. None does no conversion. 'standard' takes half of the detection limit. 'trend' is meant as an output for trend analysis with includes an additional column dtl_ratio referring to the ratio of values under the detection limit.
    output_site_data : bool
        Should the site data be output?
    sites_df : DataFrame
        The DataFrame return from the rd_hilltop_sites function. If this is passed than rd_hilltop_sites is not run.

    Returns
    -------
    DataFrame
    """

    ### First read all of the sites in the hts file and select the ones to be read
    if not isinstance(sites_df, DataFrame):
        sites_df = measurement_list(hts, sites=sites, mtypes=mtypes, rem_wq_sample=False)

    ### Select out the sites/mtypes within the date range
    if isinstance(start, str):
        sites_df = sites_df[sites_df.end_date >= start]
    if isinstance(end, str):
        sites_df = sites_df[sites_df.start_date <= end]

    ### Open the hts file
    wqr = Dispatch("Hilltop.WQRetrieval")
    dfile = Dispatch("Hilltop.DataFile")
    try:
        dfile.Open(hts)
    except ValueError:
        print(dfile.errmsg)

    ### Iterate through he hts file
    df_lst = []
    for i in sites_df.index:
        site_data = sites_df.loc[i]
        site = site_data['site']
        mtype = site_data['mtype']
        if mtype == 'WQ Sample':
            continue
        wqr = dfile.FromWQSite(site, mtype)

        ## Set up start and end times and aggregation initiation
        if start is None:
            start1 = wqr.DataStartTime
        else:
            start1 = pywintypes.TimeType.strptime(start, '%Y-%m-%d')
        if end is None:
            end1 = wqr.DataEndTime
        else:
            end1 = pywintypes.TimeType.strptime(end, '%Y-%m-%d')

        if not wqr.FromTimeRange(start1, end1):
            continue

        ## Extract data
        data = []
        time = []
        sample_p = []

        test_params = sites_df[sites_df.site == site].mtype.unique()
        if ('WQ Sample' in test_params) & (isinstance(mtype_params, list) | isinstance(sample_params, list)):
            mtype_p = []
            while wqr.GetNext:
                data.append(str(wqr.value.encode('ascii', 'ignore').decode()))
                time.append(str(pytime_to_datetime(wqr.time)))
                sample_p.append({sp: str(wqr.params(sp).encode('ascii', 'ignore').decode()) for sp in sample_params})
                mtype_p.append({mp: str(wqr.params(mp).encode('ascii', 'ignore').decode()) for mp in mtype_params})
        else:
            while wqr.GetNext:
                data.append(str(wqr.value.encode('ascii', 'ignore').decode()))
                time.append(str(pytime_to_datetime(wqr.time)))

        if data:
            df_temp = DataFrame({'time': time, 'data': data, 'site': site, 'mtype': mtype})
            if sample_p:
                df_temp = concat([df_temp, DataFrame(sample_p), DataFrame(mtype_p)], axis=1)
            df_lst.append(df_temp)

    dfile.Close()
    wqr.close()
    if df_lst:
        data = concat(df_lst)
        data.loc[:, 'time'] = to_datetime(data.loc[:, 'time'])
        data1 = to_numeric(data.loc[:, 'data'], errors='coerce')
        data.loc[data1.notnull(), 'data'] = data1[data1.notnull()]
        data = data.reset_index(drop=True)

        #### Convert detection limit values
        if dtl_method is not None:
            less1 = data['data'].str.match('<')
            if less1.sum() > 0:
                less1.loc[less1.isnull()] = False
                data2 = data.copy()
                data2.loc[less1, 'data'] = to_numeric(data.loc[less1, 'data'].str.replace('<', ''), errors='coerce') * 0.5
                if dtl_method == 'standard':
                    data3 = data2
                if dtl_method == 'trend':
                    df1 = data2.loc[less1]
                    count1 = data.groupby('mtype')['data'].count()
                    count1.name = 'tot_count'
                    count_dtl = df1.groupby('mtype')['data'].count()
                    count_dtl.name = 'dtl_count'
                    count_dtl_val = df1.groupby('mtype')['data'].nunique()
                    count_dtl_val.name = 'dtl_val_count'
                    combo1 = concat([count1, count_dtl, count_dtl_val], axis=1, join='inner')
                    combo1['dtl_ratio'] = (combo1['dtl_count'] / combo1['tot_count']).round(2)

                    ## conditionals
                    param2 = combo1[(combo1['dtl_ratio'] > 0.4) & (combo1['dtl_val_count'] != 1)]
                    over_40 = data['mtype'].isin(param2.index)

                    ## Calc detection limit values
                    data3 = merge(data, combo1['dtl_ratio'].reset_index(), on='mtype', how='left')
                    data3.loc[:, 'data_dtl'] = data2['data']

                    max_dtl_val = data2[over_40 & less1].groupby('mtype')['data'].transform('max')
                    max_dtl_val.name = 'dtl_data_max'
                    data3.loc[over_40 & less1, 'data_dtl'] = max_dtl_val
            else:
                data3 = data
        else:
            data3 = data

        if output_site_data:
            sites_df = sites_df[~(sites_df.mtype == 'WQ Sample')]
            return data3, sites_df
        else:
            return data3


def write_wq_data(hts, data):
    """
    Function to write water quality data to Hilltop hts files.

    Parameters
    ------------
    hts : str
        Path to the hts file to write to.
    data : dict
        A dictionary of the data in the following example structure
        {'SQ00001': {'2000-01-01 12:00:00': {'SiteParameter':
                                   {'Sample ID': '174759302',
                                    'Project': 'BPStreams'
                                    }
                               'Measurement':
                                   {'Conductivity (Field)':
                                        {'Value': '18.59',
                                         'MethodText': 'Unknown meter',
                                         'Lab': 'Field'
                                         }
                                        }
                                       }
                     }
         }

    Returns
    -------
    None
    """
    for s in data:
        print(s)
        dfile = Dispatch("Hilltop.WQInput")
        try:
            dfile.Open(hts)
        except ValueError:
            print(dfile.ErrorMsg)

        for d in data[s]:
            # print(d)
            q1 = data[s][d]

            dfile.PutSample(s, d) # Initialize the site and time
            if 'SiteParameter' in q1:
                sp4 = q1['SiteParameter']
                for key, val in sp4.items():
                    dfile.SetParam(key, val) # set the sample paremters, not associated with measurement types
            if 'Measurement' in q1:
                m1 = q1['Measurement']
                for key, val in m1.items():
                    dup = val.copy()
                    dfile.PutMeasurement(key, dup.pop('Value')) # Set a measurement value
                    if dup:
                        for mp, mpval in dup.items():
                            dfile.SetParam(mp, mpval) # sample parameters associated with the measurement parameter

        if len(dfile.ErrorMsg) > 0:
            raise ValueError(dfile.ErrorMsg)
        else:
            dfile.Close()   # commit data to the hts file



# -*- coding: utf-8 -*-
"""
Created on Wed May 30 12:05:46 2018

@author: MichaelEK
"""
from hilltoppy.web_service import measurement_list, site_list, get_data, wq_sample_parameter_list

### Parameters
base_url = 'http://data.ecan.govt.nz/'
hts = 'WQAll.hts'
site = 'SQ31045'
measurement = 'Total Phosphorus'
from_date = '1983-11-22 10:50'
to_date = '2018-04-13 14:05'
dtl_method = 'trend'

### Tests


def test_site_list():
    sites = site_list(base_url, hts, True)
    assert len(sites) > 9000


def test_measurement_list():
    mtype_df1 = measurement_list(base_url, hts, site)
    assert len(mtype_df1) > 30


def test_wq_sample_parameter_list():
    mtype_df2 = wq_sample_parameter_list(base_url, hts, site)
    assert len(mtype_df2) > 10


def test_get_data1():
    tsdata1 = get_data(base_url, hts, site, measurement, from_date=from_date, to_date=to_date)
    assert len(tsdata1) > 80


def test_get_data2():
    tsdata2, extra2 = get_data(base_url, hts, site, measurement, from_date=from_date, to_date=to_date, parameters=True)
    assert (len(tsdata2) > 80) & (len(extra2) > 300)


def test_get_data3():
    tsdata3 = get_data(base_url, hts, site, 'WQ Sample', from_date=from_date, to_date=to_date)
    assert len(tsdata3) > 800


def test_get_data4():
    tsdata4, extra4 = get_data(base_url, hts, site, measurement, from_date=from_date, to_date=to_date, parameters=True, dtl_method=dtl_method)
    assert (len(tsdata4) > 80) & (len(extra4) > 300) & (tsdata4.Value.dtype.name == 'float32')

